require fflib.fs
\ require debug.fs
require util.fs

library libobjc libobjc.so.1
library foundation libgnustep-base.so
library appkit libgnustep-gui.so
library libobjcwrapper ./libobjcwrapper.so.0

libobjc objc-get-class ptr (ptr) objc_get_class
libobjc objc-add-class ptr (void) __objc_add_class_to_hash
libobjc objc-lookup-class ptr (ptr) objc_lookup_class

libobjc objc-register-selectors-from-class ptr (void) __objc_register_selectors_from_class
\ libobjc objc_addMethods ptr (void) __objc_register_selectors_from_class

libobjc objc-send-msgv ptr ptr ptr (ptr) objc_msg_sendv
libobjcwrapper objc-send-msgv-wrapper ptr (ptr) objc_msg_sendv_for_forth
libobjc objc-register-name ptr (ptr) sel_register_name
libobjc objc-create-instance ptr (ptr) class_create_instance

libobjc class-get-instance-method ptr ptr (ptr) class_get_instance_method

foundation NSLog0 ptr (void) NSLog
foundation NSLog1 ptr ptr (void) NSLog

appkit NSApplicationMain int ptr (int) NSApplicationMain

require objc-runtime.fs

: objc-class ( addr u -- ptr ) ftoc objc-get-class ;
: objc-sel ( addr u -- ptr ) ftoc objc-register-name ;
\ : objc-class-add ( class -- ) objc_addClass ;

\ TODO: deal with Protocol, Category...
: objc-add-method ( class -- )
  ->meta-class @ objc-register-selectors-from-class ;

require NSObject.fs
require NSAutoreleasePool.fs
require NSString.fs
require NSMutableString.fs
require NSNumber.fs
require NSArray.fs
require NSMutableArray.fs

s" alloc" objc-sel Constant sel_alloc
s" init" objc-sel Constant sel_init
s" retain" objc-sel Constant sel_retain
s" release" objc-sel Constant sel_release
s" autorelease" objc-sel Constant sel_autorelease
s" description" objc-sel Constant sel_description
s" count" objc-sel Constant sel_count

0 [if]
\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
\ XXX"'s are only for using inside colon definition.
\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
: sel" \ usage example: sel" stringWithString:"
  POSTPONE s"
  POSTPONE ftoc
  POSTPONE objc-register-name ; immediate
: class" \ usage example: class" NSString"
  POSTPONE s"
  POSTPONE ftoc
  POSTPONE objc-get-class ; immediate

: NSStringFromForth ( addr u -- ptr )
  ftoc 1 >r >r NSString sel_stringWithCString: r> r> send-msg @ ;
immediate
: @" \ replacement for s" which produces an NSString instead
  POSTPONE s" POSTPONE NSStringFromForth ; immediate
[then]
: sel" \ usage example: sel" stringWithString:"
  state @ if
    POSTPONE s"
    POSTPONE ftoc
    POSTPONE objc-register-name
  else
    [char] " parse ftoc objc-register-name
  then ; immediate
: class" \ usage example: class" NSString"
  state @ if
    POSTPONE s"
    POSTPONE ftoc
    POSTPONE objc-get-class
  else
    [char] " parse ftoc objc-get-class
  then ; immediate

: NSStringFromForth ( addr u -- ptr )
  ftoc 1 >r >r NSString sel_stringWithCString: r> r> send-msg @ ;
: @" \ replacement for s" which produces an NSString instead
  state @ if
    POSTPONE s" POSTPONE NSStringFromForth
  else
    [char] " parse NSStringFromForth
  then ; immediate

: objc_release ( ptr -- )
  sel_release 0 send-msg drop ;
: objc_retain ( ptr -- ptr )
  sel_retain 0 send-msg drop ;
: objc_autorelease ( ptr -- ptr )
  sel_autorelease 0 send-msg @ ;

: ndrop ( x1 ... xu u -- )
  0 ?DO drop LOOP ;

1 constant CLS_CLASS
2 constant CLS_META
8 constant CLS_RESOLVED

: search-root-class ( class -- root )
  BEGIN \ search root class
    dup ->super-class @ dup 0= IF drop 1 ELSE nip 0 ENDIF
  UNTIL ;

: callocate ( u -- addr ) \ allocate and erase space
  dup allocate throw dup rot erase ;

: -@> ( "classname" superclass -- classptr methodliststruct )
  { super }
  parse-word ftoc { classname }

  \ allocate memory for class and meta structure
  objc-class-struct %size callocate { sub }
  objc-class-struct %size callocate { sub-meta }

  Object ->meta-class @ ->dtable @ { root-dtable }
  sub-meta					sub ->meta-class !
  super						sub ->super-class !
  classname					sub ->name !
  4						sub ->instance-size !
  objc-method-list-struct %size callocate	sub ->methods !
  super ->subclass-list @			sub ->sibling-class !

  sub super ->subclass-list !


  \ super search-root-class ->meta-class @ sub-meta ->meta-class !
  Object ->meta-class @				sub-meta ->meta-class !
  super ->meta-class @ 			 	sub-meta ->super-class !
  sub ->name @					sub-meta ->name !
  Object ->meta-class @ ->instance-size @	sub-meta ->instance-size !
  root-dtable				 	sub-meta ->dtable !
  super ->meta-class @ ->subclass-list @	sub-meta ->sibling-class !

  sub-meta  super ->meta-class @ ->subclass-list !

  CLS_CLASS CLS_RESOLVED or		sub ->info !
  CLS_META CLS_RESOLVED or			sub-meta ->info !

  sub objc-add-class
  sub sub ->methods @
;


: objc-method-size ( -- n )
  objc-method-struct %size ;
: objc-method-list-size ( - n )
  objc-method-list-struct %size ;
: calc-method-list-size ( list -- n )
  ->method-count @ objc-method-size *  2 cells + ;
: alloc-another-method-struct ( addr1 - addr2 )
  dup calc-method-list-size objc-method-size +  resize throw ;
: update-one-method-struct ( sel sgn cb addr - )
  >r r@ ->method-imp ! r@ ->method-types ! r> ->method-name !  ;
: update-method-list ( sel sgn cb addr - )
  dup dup calc-method-list-size + swap >r update-one-method-struct
  1 r> ->method-count +!  ;

: @method ( "name" "signature" methodliststruct callback --
methodstruct )
  parse-word objc-sel parse-word ftoc { mls cb sel sgn }
  mls alloc-another-method-struct to mls
  sel sgn cb mls update-method-list
  mls ;

: @end ( classstruct methodliststruct -- )
 over ->methods !  objc-add-method ;

Variable NSAutoreleasePoolInitDone 0 NSAutoreleasePoolInitDone !
Variable pool 0 pool !

: NSAutoreleasePool-init
  NSAutoreleasePoolInitDone @ if exit then
  NSAutoreleasePool sel_init 0 send-msg @ pool !
  1 NSAutoreleasePoolInitDone !
  CR ." Pool init done." ;

: NSAutoreleasePool-release
  NSAutoreleasePoolInitDone @ if
    pool @ sel_release 0 send-msg drop CR ." Pool release done."
    0 NSAutoreleasePoolInitDone ! exit
  then ;

callback IBAction (void) ptr ptr ptr callback; \ parameters: self _cmd sender
